local FRAMES = {
	-- default UI
	"SpellBookFrame",
	"QuestLogFrame",
	"FriendsFrame",
	"LFGParentFrame",
	"GossipFrame",
	"MerchantFrame",
	"MailFrame",
	"OpenMailFrame",
	"GuildRegistrarFrame",
	"DressUpFrame",
	"TabardFrame",
	"TaxiFrame",
	"QuestFrame",
	"TradeFrame",
	"LootFrame",
	"PetStableFrame",
	"StackSplitFrame",
	"WorldStateScoreFrame",
	"BattlefieldFrame",
	"ArenaFrame",
	"ItemTextFrame",
	-- addons
	"LudwigFrame",
	"LootLinkFrame", "LootLinkSearchFrame",
}
local CHILD_FRAMES = {
	-- default UI
	"PaperDollFrame",
	{"PetPaperDollFrameCompanionFrame", "CharacterFrame"},
	{"PetPaperDollFramePetFrame", "CharacterFrame"},
	"ReputationFrame",
	"SkillFrame",
	"SendMailFrame",
	"TokenFrame",
}
local LOD_FRAMES = {
	-- default UI
	Blizzard_AuctionUI = {
		"AuctionFrame",
	},
	Blizzard_CraftUI = {
		"CraftFrame",
	},
	Blizzard_InspectUI = {
		"InspectFrame",
	},
	Blizzard_ItemSocketingUI = {
		"ItemSocketingFrame",
	},
	Blizzard_MacroUI = {
		"MacroFrame",
	},
	Blizzard_TalentUI = {
		"PlayerTalentFrame",
	},
	Blizzard_TradeSkillUI = {
		"TradeSkillFrame",
	},
	Blizzard_TrainerUI = {
		"ClassTrainerFrame",
	},
	Blizzard_GuildBankUI = {
		"GuildBankFrame",
	},
	Blizzard_TimeManager = {
		"TimeManagerFrame",
	},
	Blizzard_Calendar = {
		"CalendarFrame",
	},
	-- addons
	DBM_GUI = {
		"DBMBossModFrame",
	},
}
local LOD_CHILD_FRAMES = {
	-- default UI
	Blizzard_GlyphUI = {
		{"GlyphFrame", "PlayerTalentFrame"},
	},
	Blizzard_AchievementUI = {
		"AchievementFrameHeader",
	},
	Blizzard_InspectUI = {
		"InspectPVPFrame",
		"InspectTalentFrame",
	},
	Blizzard_TokenUI = {
		"TokenFrame",
	},
}

-- for debugging / dev only
local debug = false
local showFrame = false

-- LOD frames that have been loaded
local loaded = {}
local childLoaded = {}

-- child to parent relationships
local relationship = {}

-- create the addon
nDragIt = LibStub("AceAddon-3.0"):NewAddon("nDragIt", "AceEvent-3.0", "AceHook-3.0")

SLASH_NFI1 = "/nfi"
SlashCmdList["NFI"] = function()
	local frameStack = nil
	local frame = GetMouseFocus()
	while frame do
		frameStack = frameStack and frameStack..", "..frame:GetName() or frame:GetName()
		frame = frame:GetParent()
	end
	
	DEFAULT_CHAT_FRAME:AddMessage(frameStack)
end

function nDragIt:OnInitialize()
	if debug then DEFAULT_CHAT_FRAME:AddMessage("nDragIt:OnInitialize") end

	for _, name in pairs(FRAMES) do
		self:ActivateFrame(name)
		loaded[name] = true
	end
	
	for _, name in pairs(CHILD_FRAMES) do
		self:ActivateChildFrame(name)
		childLoaded[name] = true
	end
	
	self:RegisterEvent("ADDON_LOADED")
end

function nDragIt:ADDON_LOADED(_, addon)
	
	local frames = LOD_FRAMES[addon]
	local frameName
	if frames then
		for _, frame in pairs(frames) do
			if type(frame) == "table" then
				frameName = frame[1]
			else
				frameName = frame
			end
			if not loaded[frameName] then
				loaded[frameName] = true
				self:ActivateFrame(frame)
			end
		end
	end
	
	frames = LOD_CHILD_FRAMES[addon]
	if frames then
		for _, frame in pairs(frames) do
			if type(frame) == "table" then
				frameName = frame[1]
			else
				frameName = frame
			end
			if not childLoaded[frameName] then
				childLoaded[frameName] = true
				self:ActivateChildFrame(frame)
			end
		end
	end
end

function nDragIt:ActivateFrame(name)
	frame = getglobal(name)
	if frame then
		frame:EnableMouse(true)
		frame:SetMovable(true)
		frame:SetClampedToScreen(false)
		self:HookScript(frame, "OnMouseDown")
		self:HookScript(frame, "OnMouseUp")
	else
		if debug then DEFAULT_CHAT_FRAME:AddMessage("nDragIt:ActivateFrame couldn't find: "..tostring(name)) end
	end
end

function nDragIt:ActivateChildFrame(name)
	local frame, parent
	local frameName, parentName
	if type(name) == "table" then
		frameName = name[1]
		parentName = name[2]
		frame = getglobal(frameName)
		parent = getglobal(parentName)
	else
		frameName = name
		frame = getglobal(frameName)
		if frame then
			parent = frame:GetParent()
			parentName = parent:GetName()
		end
	end
	if frame and parent then
		frame:EnableMouse(true)
		frame:SetClampedToScreen(false)
		parent:SetMovable(true)
		parent:SetClampedToScreen(false)
		self:HookScript(frame, "OnMouseDown", "ChildOnMouseDown")
		self:HookScript(frame, "OnMouseUp", "ChildOnMouseUp")
		relationship[frameName] = parentName
	else
		if debug then DEFAULT_CHAT_FRAME:AddMessage("nDragIt:ActivateChildFrame couldn't find: "..tostring(frameName)..","..tostring(parentName)) end
	end
end

function nDragIt:OnMouseDown(frame, button)

	if button == "LeftButton" then
		if frame then
			if showFrame then
				frame:SetBackdrop({ bgFile = "Interface/Tooltips/UI-Tooltip-Background", })
			end
			frame:StartMoving()
			frame:SetUserPlaced(false)
		end
	end
	
	self.hooks[frame].OnMouseDown(frame, button)
end

function nDragIt:OnMouseUp(frame, button)

	if frame then
		if showFrame then
			frame:SetBackdrop({})
		end
		frame:StopMovingOrSizing()
	end
	
	self.hooks[frame].OnMouseUp(frame, button)
end

function nDragIt:ChildOnMouseDown(frame, button)

	if button == "LeftButton" then
		if frame then
			parentFrame = nDragIt:ResolveParent(frame)
			if parentFrame then
				if showFrame then
					frame:SetBackdrop({ bgFile = "Interface/Tooltips/UI-Tooltip-Background", })
				end
				parentFrame:StartMoving()
				parentFrame:SetUserPlaced(false)
			end
		end
	end
	
	self.hooks[frame].OnMouseDown(frame, button)
end

function nDragIt:ChildOnMouseUp(frame, button)

	if frame then
		parentFrame = nDragIt:ResolveParent(frame)
		if parentFrame then
			if showFrame then
				frame:SetBackdrop({})
			end
			parentFrame:StopMovingOrSizing()
		end
	end
	
	self.hooks[frame].OnMouseUp(frame, button)
end

function nDragIt:ResolveParent(frame)
	local parent
	
	local frameName = frame:GetName()
	if relationship[frameName] then
		parent = getglobal(relationship[frameName]) 
	end
	
	return parent
end
